#include <iostream>
#include <string>
#include <slotsig/slotsig.h>

void f(nsSlotSig::clSenderBase* sender, int i)
{
  std::cout << "::f(" << i << "), sender = " ;
  if ( sender != 0 )
    std::cout << sender->senderName() << ", class " <<
      sender->senderClassName() << std::endl ;
  else
    std::cout << "<null>\n" ;
}

class clMySlot : public nsSlotSig::clSlot1<void, int>
{
  public:
    clMySlot() :
      nsSlotSig::clSlot1<void, int>()
    { }

    void operator() (int i)
    {
      std::cout << "clMySlot::operator() (" << i << "), sender = " ;
      if ( sender() != 0 )
        std::cout << sender()->senderName() << ", class " <<
          sender()->senderClassName() << std::endl ;
      else
        std::cout << "<null>\n" ;
    }
} ;

class clMyClass
{
  public:
    clMyClass() :
      m_Slot1(this, &clMyClass::f)
    { }

    nsSlotSig::clSlot1<void, int>* slot()
    { return &m_Slot1 ; }

  private:
    void f(int i)
    {
      std::cout << "clMyClass::f(" << i << "), sender = " ;
      if ( m_Slot1.sender() != 0 )
        std::cout << m_Slot1.sender()->senderName() << ", class " <<
          m_Slot1.sender()->senderClassName() << std::endl ;
      else
        std::cout << "<null>\n" ;
    }

    nsSlotSig::clSlotInClass1<clMyClass, void, int> m_Slot1 ;
} ;

class clMySender : public nsSlotSig::clSenderBase
{
  public:
    clMySender() :
      nsSlotSig::clSenderBase("clMySender", "instance of clMySender"),
      m_Sig(this)
    { }

    nsSlotSig::clSig1<void, int>& signal()
    { return m_Sig ; }

    void doEmit()
    { m_Sig.emit(10) ; }

  private:
    nsSlotSig::clSig1<void, int> m_Sig ;
} ;

int main(int argc, char* argv[])
{
  clMySender my_sender ;
  clMySlot my_slot ;
  clMyClass my_class ;

  my_sender.signal().connect(f) ;
  my_sender.signal().connect(&my_slot) ;
  my_sender.signal().connect(my_class.slot()) ;

  my_sender.doEmit() ;

  nsSlotSig::clSig1<void, int> sig ;
  sig.connect(f) ;
  sig.connect(&my_slot) ;
  sig.connect(my_class.slot()) ;

  sig.emit(20) ;

  return 0 ;
}
