/*
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

template <typename _RETURN_TYPE_
          __COMMA__
          __TEMPLATE_ARGS__>
class __CL_SIG_NAME__ : public __CL_SLOTSIG_BASE_NAME__<_RETURN_TYPE_
                                                        __COMMA__
                                                        __TEMPLATE_PARAMS__>
{
  friend class __CL_SLOT_NAME__<_RETURN_TYPE_
                                __COMMA__
                                __TEMPLATE_PARAMS__> ;
  public:
    typedef __CL_SLOT_NAME__<_RETURN_TYPE_
                             __COMMA__
                             __TEMPLATE_PARAMS__> clSlot ;
    typedef std::set<clSlot*> clSlotSet ;
    typedef __CL_SLOTSIG_BASE_NAME__<_RETURN_TYPE_
                                     __COMMA__
                                     __TEMPLATE_PARAMS__> clSlotSigBase ;

    typedef _RETURN_TYPE_ (*FuncType_t) (__METHOD_ARGS__) ;
    typedef _RETURN_TYPE_ (*FuncTypeSender_t) (clSenderBase* sender
                                               __COMMA__
                                               __METHOD_ARGS__) ;
    typedef std::set<FuncType_t>        clFuncSet ;
    typedef std::set<FuncTypeSender_t>  clFuncSenderSet ;

    __CL_SIG_NAME__(clSenderBase* send = 0) :
      clSlotSigBase(send),
      m_FuncSet(),
      m_FuncSenderSet()
    {
      VERB_SLOTSIG(5, VERB_NBPAR("clSig") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSig") << "(" << VERB_PTR(send) << "), instance #") ;
      ++m_NbInstances ;
      VERBOSE(5, m_NbInstances << "\n") ;
    }

    virtual ~__CL_SIG_NAME__()
    {
      VERB_SLOTSIG(5, VERB_NBPAR("clSig") << VERB_PTR(this) << "::"
        << VERB_NBPAR("~clSig") << "(), remaining instances: ") ;
      --m_NbInstances ;
      VERBOSE(5, m_NbInstances << "\n") ;
      disconnectAll() ;
      if ( m_NbInstances == 0 )
      {
      VERB_SLOTSIG(5, VERB_NBPAR("clSig") << VERB_PTR(this) << "::"
        << VERB_NBPAR("~clSig") << "() : last instance, deleting " << std::dec
          << m_AutoSlots.size() << " automatic slots...\n") ;
        std::set<clSlotBase*>::const_iterator iter = m_AutoSlots.begin() ;
        std::set<clSlotBase*>::const_iterator end = m_AutoSlots.end() ;
        while ( iter != end )
        {
          (*iter)->disconnectAll() ;
          delete *iter ;
          ++iter ;
        }
        m_AutoSlots.clear() ;
      }
    }

    template <typename _IN_CLASS_>
    class clSlotInClass : public __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                                          _RETURN_TYPE_
                                                          __COMMA__
                                                          __TEMPLATE_PARAMS__>
    {
      public:
        clSlotInClass() :
          __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                    _RETURN_TYPE_
                                    __COMMA__
                                    __TEMPLATE_PARAMS__>()
          { }

        clSlotInClass(_IN_CLASS_* in_class, typename clSlotInClass<_IN_CLASS_>::ClassMethod_t cm) :
          __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                    _RETURN_TYPE_
                                    __COMMA__
                                    __TEMPLATE_PARAMS__>(in_class, cm)
          { }

        clSlotInClass(_IN_CLASS_* in_class, typename clSlotInClass<_IN_CLASS_>::ClassMethodSender_t cms) :
          __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                    _RETURN_TYPE_
                                    __COMMA__
                                    __TEMPLATE_PARAMS__>(in_class, cms)
          { }

        clSlotInClass(_IN_CLASS_* in_class, typename clSlotInClass<_IN_CLASS_>::ClassMethodConst_t cm) :
          __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                    _RETURN_TYPE_
                                    __COMMA__
                                    __TEMPLATE_PARAMS__>(in_class, cm)
          { }

        clSlotInClass(_IN_CLASS_* in_class, typename clSlotInClass<_IN_CLASS_>::ClassMethodSenderConst_t cms) :
          __CL_SLOT_INCLASS_NAME__<_IN_CLASS_,
                                    _RETURN_TYPE_
                                    __COMMA__
                                    __TEMPLATE_PARAMS__>(in_class, cms)
          { }
    } ;

    virtual void connect (clSlot* slot_sig)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect (clSlot"
        << VERB_PTR(slot_sig) << ")\n") ;
      clSlotSigBase::connect(slot_sig) ;
    }

    virtual void connect    (FuncType_t  func)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect (FuncType_t"
        << VERB_PTR(func) << ")") ;
      if ( func != 0 )
        m_FuncSet.insert(func) ;
      else
        VERBOSE(0, ": Warning: given a null pointer.") ;
      VERBOSE(0, "\n") ;
    }

    virtual void connect    (FuncTypeSender_t funcs)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect (FuncTypeSender_t"
        << VERB_PTR(funcs) << ")") ;
      if ( funcs != 0 )
        m_FuncSenderSet.insert(funcs) ;
      else
        VERBOSE(0, ": Warning: given a null pointer.\n") ;
      VERBOSE(0, "\n") ;
    }

    template <typename _IN_CLASS_>
    void connect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethod_t cm)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethod_t"
        << VERB_PTR(cm) << ")\n") ;
      slotPoolMngt (in_class, cm, PoolConnect) ;
    }

    template <typename _IN_CLASS_>
    void connect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodSender_t cms)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodSender_t"
        << VERB_PTR(cms) << ")\n") ;
      slotPoolMngt (in_class, cms, PoolConnect) ;
    }

    template <typename _IN_CLASS_>
    void connect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodConst_t cm)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodConst_t"
        << VERB_PTR(cm) << ")\n") ;
      slotPoolMngt (in_class, cm, PoolConnect) ;
    }

    template <typename _IN_CLASS_>
    void connect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodSenderConst_t cms)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::connect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodSenderConst_t"
        << VERB_PTR(cms) << ")\n") ;
      slotPoolMngt (in_class, cms, 0) ;
    }

    virtual bool disconnect (clSlot* slot_sig)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect (clSlot"
        << VERB_PTR(slot_sig) << ")\n") ;
      return clSlotSigBase::disconnect(slot_sig) ;
    }

    virtual void disconnect (FuncType_t  func)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect (FuncType_t"
        << VERB_PTR(func) << ")") ;
      if ( func != 0 )
        m_FuncSet.erase(func) ;
      else
        VERBOSE(0, ": Warning: given a null pointer.") ;
      VERBOSE(0, "\n") ;
    }

    virtual void disconnect (FuncTypeSender_t funcs)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect (FuncTypeSender_t"
        << VERB_PTR(funcs) << ")") ;
      if ( funcs != 0 )
        m_FuncSenderSet.erase(funcs) ;
      else
        VERBOSE(0, ": Warning: given a null pointer.\n") ;
      VERBOSE(0, "\n") ;
    }

    template <typename _IN_CLASS_>
    void disconnect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethod_t cm)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethod_t"
        << VERB_PTR(cm) << ")\n") ;
      slotPoolMngt (in_class, cm, PoolDisconnect) ;
    }

    template <typename _IN_CLASS_>
    void disconnect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodSender_t cms)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodSender_t"
        << VERB_PTR(cms) << ")\n") ;
      slotPoolMngt (in_class, cms, PoolDisconnect) ;
    }

    template <typename _IN_CLASS_>
    void disconnect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodConst_t cm)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodConst_t"
        << VERB_PTR(cm) << ")\n") ;
      slotPoolMngt (in_class, cm, PoolDisconnect) ;
    }

    template <typename _IN_CLASS_>
    void disconnect (_IN_CLASS_* in_class,
                  typename clSlotInClass<_IN_CLASS_>::ClassMethodSenderConst_t cms)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnect ("
        << "_IN_CLASS_" << VERB_PTR(in_class) << ", ClassMethodSenderConst_t"
        << VERB_PTR(cms) << ")\n") ;
      slotPoolMngt (in_class, cms, PoolDisconnect) ;
    }

    virtual void disconnectAll()
    {
      VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::disconnectAll (): "
        << m_FuncSet.size() << " functions, "
        << m_FuncSenderSet.size() << " functions with sender.\n") ;
      clSlotSigBase::disconnectAll() ;
      m_FuncSet.clear() ;
      m_FuncSenderSet.clear() ;
    }

    virtual void emit(__METHOD_ARGS__)
    {
      VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (__METHOD_ARGS__)\n") ;
      emit (sender() __COMMA__ __METHOD_PARAMS__) ;
    }

    virtual void emit(register clSenderBase* sender __COMMA__ __METHOD_ARGS__)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (clSenderBase"
        << VERB_PTR(sender) << ", __METHOD_ARGS__)\n") ;
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to slots...\n") ;
        register typename std::set<clSlotSigBase*>::const_iterator slot_iter = slotSigSet().begin() ;
        const typename std::set<clSlotSigBase*>::const_iterator end_iter = slotSigSet().end() ;
        clSlot* slot ;
        while ( slot_iter != end_iter )
        {
          slot = dynamic_cast<clSlot*>(*slot_iter) ;
          if ( slot != 0 )
          {
            slot->setSender(sender) ;
            slot->operator() (__METHOD_PARAMS__) ;
          }
          ++slot_iter ;
        }
      }
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to functions...\n") ;
        register typename clFuncSet::const_iterator func_iter = m_FuncSet.begin() ;
        const typename clFuncSet::const_iterator end_iter = m_FuncSet.end() ;
        while ( func_iter != end_iter )
          (*(func_iter++))(__METHOD_PARAMS__) ;
      }
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to functions with sender...\n") ;
        register typename clFuncSenderSet::const_iterator func_sender_iter = m_FuncSenderSet.begin() ;
        const typename clFuncSenderSet::const_iterator end_sender_iter = m_FuncSenderSet.end() ;
        while ( func_sender_iter != end_sender_iter )
          (*(func_sender_iter++))(sender __COMMA__ __METHOD_PARAMS__) ;
      }
    }

    template <typename _ACCUMULATOR_TYPE_>
    void emit(_ACCUMULATOR_TYPE_& accum __COMMA__ __METHOD_ARGS__)
    {
      VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this)
        << "::emit (_ACCUMULATOR_, __METHOD_ARGS__)\n") ;
      emit (accum, sender() __COMMA__ __METHOD_PARAMS__) ;
    }

    template <typename _ACCUMULATOR_TYPE_>
    void emit(_ACCUMULATOR_TYPE_& accum, clSenderBase* sender __COMMA__ __METHOD_ARGS__)
    {
      VERB_SLOTSIG(0, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (_ACCUMULATOR_, clSenderBase"
        << VERB_PTR(sender) << ", __METHOD_ARGS__)\n") ;
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (_ACCUMULATOR_, clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to slots...\n") ;
        register typename std::set<clSlotSigBase*>::const_iterator slot_iter = slotSigSet().begin() ;
        const typename std::set<clSlotSigBase*>::const_iterator end_iter = slotSigSet().end() ;
        clSlot* slot ;
        while ( slot_iter != end_iter )
        {
          slot = dynamic_cast<clSlot*>(*slot_iter) ;
          if ( slot != 0 )
          {
            slot->setSender(sender) ;
            accum(slot->operator() (__METHOD_PARAMS__)) ;
          }
          ++slot_iter ;
        }
      }
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (_ACCUMULATOR_, clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to functions...\n") ;
        register typename clFuncSet::const_iterator func_iter = m_FuncSet.begin() ;
        const typename clFuncSet::const_iterator end_iter = m_FuncSet.end() ;
        while ( func_iter != end_iter )
          accum((*(func_iter++))(__METHOD_PARAMS__)) ;
      }
      {
        VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::emit (_ACCUMULATOR_, clSenderBase"
          << VERB_PTR(sender) << ", __METHOD_ARGS__): emitting to functions with sender...\n") ;
        register typename clFuncSenderSet::const_iterator func_sender_iter = m_FuncSenderSet.begin() ;
        const typename clFuncSenderSet::const_iterator end_sender_iter = m_FuncSenderSet.end() ;
        while ( func_sender_iter != end_sender_iter )
          accum((*(func_sender_iter++))(sender __COMMA__ __METHOD_PARAMS__)) ;
      }
    }

    virtual void operator() (__METHOD_ARGS__)
    {
      VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this)
        << "::operator() (__METHOD_ARGS__)n") ;
      emit(__METHOD_PARAMS__) ;
    }

    virtual void operator() (clSenderBase* send __COMMA__ __METHOD_ARGS__)
    {
      VERB_SLOTSIG(1, VERB_NBPAR("clSig") << VERB_PTR(this) << "::operator() (clSenderBase"
        << VERB_PTR(send) << ", __METHOD_ARGS__)n") ;
      emit(send __COMMA__ __METHOD_PARAMS__) ;
    }

    const clFuncSet&        funcSet() const
    { return m_FuncSet ; }

    const clFuncSenderSet&  funcSenderSet() const
    { return m_FuncSenderSer ; }

  protected:

    template <typename _IN_CLASS_, typename _METHOD_TYPE_>
    struct clMethodToSlot
    {
      clMethodToSlot (_METHOD_TYPE_ m, clSlotInClass<_IN_CLASS_>* s) :
        m_Method(m),
        m_Slot(s)
      { }
      _METHOD_TYPE_ m_Method ;
      clSlotInClass<_IN_CLASS_>* m_Slot ;
      bool operator== (_METHOD_TYPE_ m)
      { return (m_Method == m) ; }
    } ;

    typedef enum
    {
      PoolConnect,
      PoolDisconnect,
    } SlotsPoolDoWhat_t ;

    template <typename _IN_CLASS_, typename _METHOD_TYPE_>
    void slotPoolMngt (_IN_CLASS_* in_class, _METHOD_TYPE_ method, SlotsPoolDoWhat_t do_what)
    {
      typedef std::list<clMethodToSlot<_IN_CLASS_, _METHOD_TYPE_> > clMethodsToSlots ;
      typedef std::map<_IN_CLASS_*, clMethodsToSlots> clSlotsPool ;
      static clSlotsPool slots_pool ;

      VERB_SLOTSIG(3, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(_IN_CLASS_"
        << VERB_PTR(in_class) << ", method" << VERB_PTR(method) << ", ") ;

      switch ( do_what )
      {
        case PoolConnect :
        {
          VERBOSE(3, "PoolConnect)\n") ;
          if ( (in_class != 0) && (method != 0) )
          {
            clSlotInClass<_IN_CLASS_>* the_slot = 0 ;
            typename clSlotsPool::iterator cl_iter = slots_pool.find(in_class) ;
            if ( cl_iter != slots_pool.end() )
            {
              VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                << "Found the given class" << VERB_PTR(in_class) << "\n") ;
              clMethodsToSlots& meth_slots = cl_iter->second ;
              typename clMethodsToSlots::const_iterator meth_iter =
                std::find(meth_slots.begin(), meth_slots.end(), method) ;
              if ( meth_iter != meth_slots.end() )
              {
                VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                  << "Found the given method" << VERB_PTR(method) << "\n") ;
                the_slot = (*meth_iter).m_Slot ;
              }
            }
            if ( the_slot == 0 )
            {
              VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                << "Creating an automatic slot, storing to the pool...\n") ;
              the_slot = new clSlotInClass<_IN_CLASS_>(in_class, method) ;
              slots_pool[in_class].push_back(clMethodToSlot<_IN_CLASS_, _METHOD_TYPE_>(method, the_slot)) ;
              clHasSlotsBase* has_slots = dynamic_cast<clHasSlotsBase*>(in_class) ;
              if ( has_slots != 0 )
                has_slots->addSlot(the_slot) ;
              else
              {
                VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                  << "Warning: _IN_CLASS_" << VERB_PTR(in_class)
                  << " does not inherit from clHasSlotsBase.\n") ;
                m_AutoSlots.insert(the_slot) ;
              }
            }
            connect (the_slot) ;
          }
          else
            VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
              << "Warning: given null class or method.\n") ;
        }
        break ;
        case PoolDisconnect :
        {
          VERBOSE(3, "PoolDisconnect)\n") ;

          if ( (in_class != 0) && (method != 0) )
          {
            clSlotInClass<_IN_CLASS_>* the_slot = 0 ;
            typename clSlotsPool::iterator cl_iter = slots_pool.find(in_class) ;
            if ( cl_iter != slots_pool.end() )
            {
              VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                << "Found the given class" << VERB_PTR(in_class) << "\n") ;
              clMethodsToSlots& meth_slots = cl_iter->second ;
              typename clMethodsToSlots::iterator meth_iter =
                std::find(meth_slots.begin(), meth_slots.end(), method) ;
              if ( meth_iter != meth_slots.end() )
              {
                VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                  << "Found the given method" << VERB_PTR(method) << "\n") ;
                the_slot = (*meth_iter).m_Slot ;
                disconnect(the_slot) ;
                clHasSlotsBase* has_slots = dynamic_cast<clHasSlotsBase*>(in_class) ;
                if ( has_slots == 0 )
                {
                  VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                    << "Warning: _IN_CLASS_" << VERB_PTR(in_class)
                    << " does not inherit from clHasSlotsBase.\n") ;
                  if ( the_slot->sigSet().size() == 0 )
                  {
                    VERB_SLOTSIG(4, VERB_NBPAR("clSig") << VERB_PTR(this)
                      << "::slotPoolMngt(): "
                      << "Nothing is connected to this slot" << VERB_PTR(the_slot)
                      << ", deleting it...\n") ;
                    m_AutoSlots.erase(the_slot) ;
                    delete the_slot ;
                    meth_slots.erase(meth_iter) ;
                  }
                }
              }
              else
                VERB_SLOTSIG(3, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                  << "Warning: given method " << VERB_PTR(method) << " not found.\n") ;
            }
            else
              VERB_SLOTSIG(3, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
                << "Warning: given class" << VERB_PTR(in_class) << "not found.\n") ;
          }
          else
            VERB_SLOTSIG(3, VERB_NBPAR("clSig") << VERB_PTR(this) << "::slotPoolMngt(): "
              << "Warning: given null class or method.\n") ;
        }
        break ;
      } // switch ( do_what )
    }

  private:

    static unsigned m_NbInstances ;
    static std::set<clSlotBase*> m_AutoSlots ;
    clFuncSet       m_FuncSet ;
    clFuncSenderSet m_FuncSenderSet ;

} ; // class clSig0
template <typename _RETURN_TYPE_
          __COMMA__
          __TEMPLATE_ARGS__>
unsigned __CL_SIG_NAME__<_RETURN_TYPE_ __COMMA__ __TEMPLATE_PARAMS__>::m_NbInstances = 0 ;

template <typename _RETURN_TYPE_
          __COMMA__
          __TEMPLATE_ARGS__>
std::set<clSlotBase*>
 __CL_SIG_NAME__<_RETURN_TYPE_ __COMMA__ __TEMPLATE_PARAMS__>::m_AutoSlots ;
