/*
 * Part of the SlogSig project, an attempt to provide a type-safe
 * signals/slots mechanism for the C++ language.
 *
 * Copyright (C) 2003 Yves Bailly <kafka_AT_kafka-fr_DOT_net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
*/

template <typename _IN_CLASS_,
          typename _RETURN_TYPE_
          __COMMA__
          __TEMPLATE_ARGS__>
class __CL_SLOT_INCLASS_NAME__ : public __CL_SLOT_NAME__<_RETURN_TYPE_
                                                          __COMMA__
                                                         __TEMPLATE_PARAMS__>
{
  friend class __CL_SIG_NAME__<_RETURN_TYPE_
                               __COMMA__
                               __TEMPLATE_PARAMS__> ;
  public:
    typedef _RETURN_TYPE_ (_IN_CLASS_::* ClassMethod_t) (__METHOD_ARGS__) ;
    typedef _RETURN_TYPE_ (_IN_CLASS_::* ClassMethodSender_t) (clSenderBase *
                                                               __COMMA__
                                                               __METHOD_ARGS__) ;
    typedef _RETURN_TYPE_ (_IN_CLASS_::* ClassMethodConst_t) (__METHOD_ARGS__) const ;
    typedef _RETURN_TYPE_ (_IN_CLASS_::* ClassMethodSenderConst_t) (clSenderBase *
                                                                    __COMMA__
                                                                    __METHOD_ARGS__) const ;
    typedef __CL_SLOT_NAME__<_RETURN_TYPE_
                             __COMMA__
                             __TEMPLATE_PARAMS__> clSlotBase ;

    __CL_SLOT_INCLASS_NAME__() :
      clSlotBase(),
      m_InClass(0),
      m_MethodType(NoMethod)
    {
      VERB_SLOTSIG(4, VERB_NBPAR("clSlotInClass") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSlotInClass") << "()\n") ;
    }

    __CL_SLOT_INCLASS_NAME__ (_IN_CLASS_* in_class, ClassMethod_t cm = 0) :
      clSlotBase(),
      m_InClass(in_class),
      m_MethodType(Method),
      m_ClassMethod(cm)
    {
      VERB_SLOTSIG(4, VERB_NBPAR("clSlotInClass") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSlotInClass") << "(_IN_CLASS_"
        << VERB_PTR(in_class) << ", ClassMethod_t" << VERB_PTR(cm) << ")\n") ;
    }

    __CL_SLOT_INCLASS_NAME__ (_IN_CLASS_* in_class, ClassMethodSender_t cms = 0) :
      clSlotBase(),
      m_InClass(in_class),
      m_MethodType(MethodSender),
      m_ClassMethodSender(cms)
    {
      VERB_SLOTSIG(4, VERB_NBPAR("clSlotInClass") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSlotInClass") << "(_IN_CLASS_"
        << VERB_PTR(in_class) << ", ClassMethodSender_t" << VERB_PTR(cms) << ")\n") ;
    }

    __CL_SLOT_INCLASS_NAME__ (_IN_CLASS_* in_class, ClassMethodConst_t cmc = 0) :
      clSlotBase(),
      m_InClass(in_class),
      m_MethodType(MethodConst),
      m_ClassMethodConst(cmc)
    {
      VERB_SLOTSIG(4, VERB_NBPAR("clSlotInClass") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSlotInClass") << "(_IN_CLASS_"
        << VERB_PTR(in_class) << ", ClassMethodConst_t" << VERB_PTR(cmc) << ")\n") ;
    }

    __CL_SLOT_INCLASS_NAME__ (_IN_CLASS_* in_class, ClassMethodSenderConst_t cmsc = 0) :
      clSlotBase(),
      m_InClass(in_class),
      m_MethodType(MethodSenderConst),
      m_ClassMethodSenderConst(cmsc)
    {
      VERB_SLOTSIG(4, VERB_NBPAR("clSlotInClass") << VERB_PTR(this) << "::"
        << VERB_NBPAR("clSlotInClass") << "(_IN_CLASS_"
        << VERB_PTR(in_class) << ", ClassMethodSenderConst_t" << VERB_PTR(cmsc) << ")\n") ;
    }

    void setInClass (_IN_CLASS_* in_class)
    { m_InClass = in_class ; }

    _IN_CLASS_* inClass() const
    { return m_InClass ; }

    virtual _RETURN_TYPE_ operator() (__METHOD_ARGS__)
    {
      VERB_SLOTSIG(3, VERB_NBPAR("clSlotInClass") << VERB_PTR(this)
        << "::operator() (__METHOD_ARGS__)\n") ;
      if ( m_InClass != 0 )
      {
        switch ( m_MethodType )
        {
          case Method :
            return (m_InClass->*m_ClassMethod)(__METHOD_PARAMS__) ;
            break ;
          case MethodSender :
            return (m_InClass->*m_ClassMethodSender)(sender() __COMMA__ __METHOD_PARAMS__) ;
            break ;
          case MethodConst :
            return (m_InClass->*m_ClassMethodConst)(__METHOD_PARAMS__) ;
            break ;
          case MethodSenderConst :
            return (m_InClass->*m_ClassMethodSenderConst)(sender() __COMMA__ __METHOD_PARAMS__) ;
            break ;
          case NoMethod:
            return clSlotBase::operator() (__METHOD_PARAMS__) ;
            break ;
        } // switch ( m_MethodType )
        return clSlotBase::operator() (__METHOD_PARAMS__) ;
      }
      return clSlotBase::operator() (__METHOD_PARAMS__) ;
    }

  private:
    _IN_CLASS_  *       m_InClass ;
    SlotMethodType_t    m_MethodType ;
    union
    {
      ClassMethod_t             m_ClassMethod ;
      ClassMethodSender_t       m_ClassMethodSender ;
      ClassMethodConst_t        m_ClassMethodConst ;
      ClassMethodSenderConst_t  m_ClassMethodSenderConst ;
    } ;
} ;
