#include <iostream>

//#define __SLOTSIG_VERBOSE__
#define __SLOTSIG_VERBOSE_LEVEL__ 10
//#define __SLOTSIG_VERBOSE_OUTPUT__ std::cerr
#include "../slotsig/slotsig.h"

#define VERBOSE(str) { std::cout << ">> TEST << " << str << std::flush ; }
#define VERB_PTR(ptr) "[" << std::hex << ptr << "]" << std::dec
#define VERB_CLASS(cl, str) { std::cout << ">> TEST << " << cl << VERB_PTR(this) << "::" \
  << str << std::flush ; }
#define VERB_SENDER(send) VERB_PTR(send) << "{\"" << \
  ((send!=0)?send->senderClassName():"") \
  << "\", \"" << ((send!=0)?send->senderName():"") << "\"}"

void f_void_int (int i)
{
  VERBOSE("::f_void_int(" << std::dec << i << ")\n") ;
}

void f_void_sender_int (nsSlotSig::clSenderBase* sender, int i)
{
  VERBOSE("::f_void_sender_int(" << VERB_SENDER(sender) << ", "
    << i << ")\n") ;
}

int f_int_int_int (int a, int b)
{
  VERBOSE("::f_int_int_int("
    << std::dec << a << ", "
    << std::dec << b << ") --> "
    << std::dec << a+b << std::endl) ;
  return a+b ;
}

int f_int_sender_int_int (nsSlotSig::clSenderBase* sender, int a, int b)
{
  VERBOSE("::f_int_int_int(" << VERB_SENDER(sender) << ", "
    << std::dec << a << ", "
    << std::dec << b << ") --> "
    << std::dec << a+b << std::endl) ;
  return a+b ;
}

typedef nsSlotSig::clSlot1<void, int> clSlotVoidIntBase ;
typedef nsSlotSig::clSlot2<int, int, int> clSlotIntIntIntBase ;

class clSlotVoidInt : public clSlotVoidIntBase
{
  public:
    clSlotVoidInt() :
      clSlotVoidIntBase()
    {
      VERB_CLASS("clSlotVoidInt", "clSlotVoidInt()\n") ;
    }

    void operator() (int i)
    {
      VERB_CLASS("clSlotVoidInt", "operator() (" << std::dec << i << "), sender = "
        << VERB_SENDER(sender()) << std::endl) ;
    }
} ;

class clSlotIntIntInt : public clSlotIntIntIntBase
{
  public:
    clSlotIntIntInt() :
      clSlotIntIntIntBase()
    {
      VERB_CLASS("clSlotIntIntInt", "clSlotIntIntInt()\n") ;
    }

    int operator() (int a, int b)
    {
      VERB_CLASS("clSlotIntIntInt", "operator() ("
        << std::dec << a << ", "
        << std::dec << b << "), sender = "
        << VERB_SENDER(sender())
        << " --> " << std::dec << a+b << std::endl) ;
      return a+b ;
    }
} ;

class clAnyClass
{
  public:
    clAnyClass()
    {
      VERB_CLASS("clAnyClass", "clAnyClass()\n") ;
    }

    virtual ~clAnyClass() { }

    void meth_void_int (int i)
    {
      VERB_CLASS("clAnyClass", "meth_void_int(" << std::dec << i << ")\n") ;
    }

    void meth_void_sender_int (nsSlotSig::clSenderBase* sender, int i)
    {
      VERB_CLASS("clAnyClass", "meth_void_sender_int("
        << VERB_SENDER(sender) << ", "
        << std::dec << i << ")\n") ;
    }

    int meth_int_int_int (int a, int b)
    {
      VERB_CLASS("clAnyClass", "meth_int_int_int("
        << std::dec << a << ", "
        << std::dec << b << ") --> "
        << std::dec << a+b << std::endl) ;
      return a+b ;
    }

    int meth_int_sender_int_int (nsSlotSig::clSenderBase* sender, int a, int b)
    {
      VERB_CLASS("clAnyClass", "meth_int_int_int("
        << VERB_SENDER(sender) << ", "
        << std::dec << a << ", "
        << std::dec << b << ") --> "
        << std::dec << a+b << std::endl) ;
      return a+b ;
    }
} ;

class clHavingSlots : public nsSlotSig::clHasSlotsBase
{
  public:
    clHavingSlots() :
      nsSlotSig::clHasSlotsBase()
    {
      VERB_CLASS("clHavingSlots", "clHavingSlots()\n") ;
    }

    void meth_void_int (int i)
    {
      VERB_CLASS("clHavingSlots", "meth_void_int(" << std::dec << i << ")\n") ;
    }

    void meth_void_sender_int (nsSlotSig::clSenderBase* sender, int i)
    {
      VERB_CLASS("clHavingSlots", "meth_void_sender_int("
        << VERB_SENDER(sender) << ", "
        << std::dec << i << ")\n") ;
    }

    int meth_int_int_int (int a, int b)
    {
      VERB_CLASS("clHavingSlots", "meth_int_int_int("
        << std::dec << a << ", "
        << std::dec << b << ") --> "
        << std::dec << a+b << std::endl) ;
      return a+b ;
    }

    int meth_int_sender_int_int (nsSlotSig::clSenderBase* sender, int a, int b)
    {
      VERB_CLASS("clHavingSlots", "meth_int_int_int("
        << VERB_SENDER(sender) << ", "
        << std::dec << a << ", "
        << std::dec << b << ") --> "
        << std::dec << a+b << std::endl) ;
      return a+b ;
    }
} ;

typedef nsSlotSig::clSig1<void, int> clSigVoidIntBase ;
typedef nsSlotSig::clSig2<int, int, int> clSigIntIntIntBase ;

class clDefSlots
{
  public:
    clDefSlots() :
      m_SlotVoidInt(this, &clDefSlots::meth_void_int),
      m_SlotIntIntInt(this, &clDefSlots::meth_int_int_int)
    {
      VERB_CLASS("clDefSlots", "clDefSlots()\n") ;
    }

    void meth_void_int(int i)
    {
      VERB_CLASS("clDefSlots", "meth_void_int(" << std::dec << i << ")\n") ;
    }

    int meth_int_int_int (int a, int b)
    {
      VERB_CLASS("clDefSlots", "meth_int_int_int("
        << std::dec << a << ", "
        << std::dec << b << ") --> "
        << std::dec << a+b << std::endl) ;
      return a+b ;
    }

    clSlotVoidIntBase* slot_void_int()
    { return &m_SlotVoidInt ; }

    clSlotIntIntIntBase* slot_int_int_int()
    { return &m_SlotIntIntInt ; }

  private:
    nsSlotSig::clSlotInClass1<clDefSlots, void, int> m_SlotVoidInt ;
    nsSlotSig::clSlotInClass2<clDefSlots, int, int, int> m_SlotIntIntInt ;
} ;

class clDefSignals : public nsSlotSig::clSenderBase
{
  public:
    explicit clDefSignals (const std::string& name) :
      nsSlotSig::clSenderBase("clDefSignals", name),
      m_SigVoidInt(this),
      m_SigIntIntInt(this)
    { }

    clSigVoidIntBase* sig_void_int()
    { return &m_SigVoidInt ; }

    clSigIntIntIntBase* sig_int_int_int()
    { return &m_SigIntIntInt ; }

    void emitAll()
    {
      m_SigVoidInt.emit(33) ;
      m_SigIntIntInt.emit(44, 55) ;
    }

  private:
    clSigVoidIntBase m_SigVoidInt ;
    clSigIntIntIntBase m_SigIntIntInt ;
} ;

class clFullClass : public nsSlotSig::clSenderBase, public nsSlotSig::clHasSlotsBase
{
} ;

int main (int argc, char* argv[])
{
  VERBOSE("-- Creating sig_void_int...\n") ;
  clSigVoidIntBase sig_void_int ;
  VERBOSE("-- Creating sig_void_int_2...\n") ;
  clSigVoidIntBase sig_void_int_2 ;
  VERBOSE("-- Creating sig_int_int_int...\n") ;
  clSigIntIntIntBase sig_int_int_int ;
  VERBOSE("-- Creating slot_void_int...\n") ;
  clSlotVoidInt slot_void_int ;
  VERBOSE("-- Creating slot_int_int_int...\n") ;
  clSlotIntIntInt slot_int_int_int ;
  VERBOSE("-- Creating any_class...\n") ;
  clAnyClass any_class ;
  VERBOSE("-- Creating having_slots...\n") ;
  clHavingSlots having_slots ;
  VERBOSE("-- Creating def_slots...\n") ;
  clDefSlots def_slots ;
  VERBOSE("-- Creating def_slots_2...\n") ;
  clDefSlots def_slots_2 ;
  VERBOSE("-- Creating def_sig...\n") ;
  clDefSignals def_sig("def_sig") ;
  VERBOSE("-- Creating def_sig_2...\n") ;
  clDefSignals def_sig_2("def_sig_2") ;

  VERBOSE("-- Connecting f_void_int to sig_void_int...\n") ;
  sig_void_int.connect(f_void_int) ;
  VERBOSE("-- Connecting slot_void_int to sig_void_int...\n") ;
  sig_void_int.connect(&slot_void_int) ;
  VERBOSE("-- Connecting any_class::meth_void_int to sig_void_int...\n") ;
  sig_void_int.connect(&any_class, &clAnyClass::meth_void_int) ;
  VERBOSE("-- Connecting having_slots::meth_void_int to sig_void_int...\n") ;
  sig_void_int.connect(&having_slots, &clHavingSlots::meth_void_int) ;
  VERBOSE("-- Connecting def_slots::slot_void_int...\n") ;
  sig_void_int.connect(def_slots.slot_void_int()) ;
  VERBOSE("-- Connecting def_slots_2::slot_void_int...\n") ;
  sig_void_int.connect(def_slots_2.slot_void_int()) ;

  VERBOSE("-- Emitting sig_void_int(77)...\n") ;
  sig_void_int.emit(77) ;

  VERBOSE("-- Connecting f_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.connect(f_void_int) ;
  VERBOSE("-- Connecting slot_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.connect(&slot_void_int) ;
  VERBOSE("-- Connecting any_class::meth_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.connect(&any_class, &clAnyClass::meth_void_int) ;
  VERBOSE("-- Connecting having_slots::meth_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.connect(&having_slots, &clHavingSlots::meth_void_int) ;
  VERBOSE("-- Connecting def_slots::slot_void_int...\n") ;
  sig_void_int_2.connect(def_slots.slot_void_int()) ;
  VERBOSE("-- Connecting def_slots_2::slot_void_int...\n") ;
  sig_void_int_2.connect(def_slots_2.slot_void_int()) ;

  VERBOSE("-- Emitting sig_void_int_2(11)...\n") ;
  sig_void_int_2.emit(11) ;

  VERBOSE("-- Disconnecting f_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.disconnect(f_void_int) ;
  VERBOSE("-- Disconnecting slot_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.disconnect(&slot_void_int) ;
  VERBOSE("-- Disconnecting any_class::meth_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.disconnect(&any_class, &clAnyClass::meth_void_int) ;
  VERBOSE("-- Disconnecting having_slots::meth_void_int to sig_void_int_2...\n") ;
  sig_void_int_2.disconnect(&having_slots, &clHavingSlots::meth_void_int) ;
  VERBOSE("-- Disconnecting def_slots::slot_void_int...\n") ;
  sig_void_int_2.disconnect(def_slots.slot_void_int()) ;
  VERBOSE("-- Disconnecting def_slots_2::slot_void_int...\n") ;
  sig_void_int_2.disconnect(def_slots_2.slot_void_int()) ;

  VERBOSE("-- Emitting sig_void_int(77)...\n") ;
  sig_void_int.emit(77) ;

  VERBOSE("-- Emitting sig_void_int_2(11)...\n") ;
  sig_void_int_2.emit(11) ;

  VERBOSE("-- Connecting def_sig::sig_void_int to def_slots::slot_void_int...\n") ;
  def_sig.sig_void_int()->connect(def_slots.slot_void_int()) ;
  VERBOSE("-- Connecting def_sig::sig_void_int to f_void_int...\n") ;
  def_sig.sig_void_int()->connect(f_void_int) ;
  VERBOSE("-- Emitting all from def_sig...\n") ;
  def_sig.emitAll() ;

  VERBOSE("-- Connecting slot_void_int to def_sig_2::sig_void_int...\n") ;
  slot_void_int.connect(def_sig_2.sig_void_int()) ;
  VERBOSE("-- Connecting def_slots::slot_void_int to def_sig_2::sig_void_int...\n") ;
  def_slots.slot_void_int()->connect(def_sig_2.sig_void_int()) ;
  VERBOSE("-- Emitting all from def_sig_2...\n") ;
  def_sig_2.emitAll() ;

  VERBOSE("-- Disconnecting def_slots::slot_void_int from def_sig::sig_void_int...\n") ;
  def_slots.slot_void_int()->disconnect(def_sig.sig_void_int()) ;
  VERBOSE("-- Emitting all from def_sig...\n") ;
  def_sig.emitAll() ;

  VERBOSE("-- Disconnecting slot_void_int to def_sig_2::sig_void_int...\n") ;
  def_sig_2.sig_void_int()->disconnect(&slot_void_int) ;
  VERBOSE("-- Disconnecting def_sig_2::sig_void_int from def_slots::slot_void_int ...\n") ;
  def_sig_2.sig_void_int()->disconnect(def_slots.slot_void_int()) ;
  VERBOSE("-- Emitting all from def_sig_2...\n") ;
  def_sig_2.emitAll() ;

  VERBOSE ("-- END TEST PROGRAM --\n") ;
  return 0 ;
}
